#include "stm32f10x.h"
#include "usart/usart_bus.h"
#include "modbus/modbus_rtu.h"
#include "counter/counter.h"
#include "utils/net.h"
#include <stdint.h>
#include <stdbool.h>
#include <stddef.h>
#include <string.h>

//! Секундный счётчик.
static counter_t counter = 0;
/**
 * Регистр хранения Modbus,
 * инкрементируется каждую секунду.
 */
static uint16_t cnt_reg = 0;

//! Шина USART.
static usart_bus_t usart_bus;
//! Modbus.
static modbus_rtu_t modbus;
/**
 * Сообщения для приёма и передачи Modbus.
 * Можно использовать и одно сообщение, если не
 * планируется обращаться к полученному сообщению
 * при формировании ответа.
 */
static modbus_rtu_message_t modbus_rx_msg, modbus_tx_msg;

//! Адрес регистра флагов со светодиодом.
#define LED_ADDRESS 0x1
//! Адрес регистра хранения со счётчиком.
#define CNT_ADDRESS 0x1

//! Порт светодиода.
#define LED_GPIO GPIOA
//! Пин светодиода.
#define LED0 GPIO_Pin_1
//! Все светодиоды для инициализации.
#define LED_ALL LED0

/**
 * Инициализирует светодиоды.
 */
static void init_leds(void)
{
    GPIO_InitTypeDef gpio_led =
        {.GPIO_Pin = LED_ALL, .GPIO_Speed = GPIO_Speed_10MHz, .GPIO_Mode = GPIO_Mode_Out_PP};
    
    GPIO_Init(GPIOA, &gpio_led);
    LED_GPIO->ODR |= LED_ALL;
}

/**
 * Выключает светодиод.
 * @param led Светодиод.
 */
static void led_off(uint16_t led)
{
    LED_GPIO->ODR |= led;
}

/**
 * Включает светодиод.
 * @param led Светодиод.
 */
static void led_on(uint16_t led)
{
    LED_GPIO->ODR &= ~led;
}

/**
 * Меняет состояние светодиода.
 * @param led Светодиод.
 */
static void led_toggle(uint16_t led)
{
    LED_GPIO->ODR ^= led;
}

/**
 * Получает состояние светодиода.
 * @param led Светодиод.
 * @return Состояние светодиода.
 */
static bool led_status(uint16_t led)
{
    return !(LED_GPIO->IDR & led);
}

/*
 * Функции для работы с USART.
 */

/**
 * Каллбэк получения байта.
 * @param byte Байт.
 * @return Флаг обработки.
 */
static bool usart_rx_byte_callback(uint8_t byte)
{
    return modbus_rtu_usart_rx_byte_callback(&modbus, byte);
}

/**
 * Каллбэк получения сообщения.
 * @return Флаг обработки.
 */
static bool usart_rx_callback(void)
{
    return modbus_rtu_usart_rx_callback(&modbus);
}

/**
 * Каллбэк отправки сообщения.
 * @return Флаг обработки.
 */
static bool usart_tx_callback(void)
{
    return modbus_rtu_usart_tx_callback(&modbus);
}

/**
 * Обработчик прерывания USART.
 */
void USART1_IRQHandler(void)
{
    usart_bus_irq_handler(&usart_bus);
}

/**
 * Обработчик прерывания канала DMA передачи данных USART.
 */
void DMA1_Channel4_IRQHandler(void)
{
    usart_bus_dma_tx_channel_irq_handler(&usart_bus);
}

/**
 * Обработчик прерывания канала DMA приёма данных USART.
 */
void DMA1_Channel5_IRQHandler(void)
{
    usart_bus_dma_rx_channel_irq_handler(&usart_bus);
}

/**
 * Обработчик прерывания системного счётчика.
 */
void SysTick_Handler(void)
{
    system_counter_tick();
}

/**
 * Включает тактирование нужной периферии.
 */
static void init_periph_clock(void)
{
    // AFIO.
    RCC_APB2PeriphClockCmd(RCC_APB2Periph_AFIO, ENABLE);
    // USART.
    RCC_APB2PeriphClockCmd(RCC_APB2Periph_GPIOA, ENABLE);
    RCC_APB2PeriphClockCmd(RCC_APB2Periph_USART1, ENABLE);
    // DMA.
    RCC_AHBPeriphClockCmd(RCC_AHBPeriph_DMA1, ENABLE);
}

/**
 * Инициализирует USART.
 */
static void init_usart(void)
{
    // Ножки GPIO.
    GPIO_InitTypeDef gpio_tx =
        {.GPIO_Pin = GPIO_Pin_9, .GPIO_Speed = GPIO_Speed_10MHz, .GPIO_Mode = GPIO_Mode_AF_PP};
    GPIO_InitTypeDef gpio_rx =
        {.GPIO_Pin = GPIO_Pin_10, .GPIO_Speed = GPIO_Speed_10MHz, .GPIO_Mode = GPIO_Mode_IN_FLOATING};
    GPIO_Init(GPIOA, &gpio_tx);
    GPIO_Init(GPIOA, &gpio_rx);
    
    // USART1: 115200 b/s, 8 бит данных, контроль чётности - "чётный", 1 стоп бит.
    USART_InitTypeDef usart_is =
        {.USART_BaudRate = 115200, .USART_WordLength = USART_WordLength_9b, .USART_StopBits = USART_StopBits_1,
         .USART_Parity = USART_Parity_Even, .USART_Mode = USART_Mode_Rx | USART_Mode_Tx,
         .USART_HardwareFlowControl = USART_HardwareFlowControl_None};
    USART_Init(USART1, &usart_is);
    USART_Cmd(USART1, ENABLE);
    
    // Шина USART.
    usart_bus_init_t usartb_is = {
        .usart_device = USART1, // Периферия.
        .dma_tx_channel = DMA1_Channel4, // Канал DMA на передачу.
        .dma_rx_channel = DMA1_Channel5 // Канал DMA на приём.
    };
    usart_bus_init(&usart_bus, &usartb_is);
    
    // Установка каллбэков.
    usart_bus_set_rx_callback(&usart_bus, usart_rx_callback);
    usart_bus_set_tx_callback(&usart_bus, usart_tx_callback);
    usart_bus_set_rx_byte_callback(&usart_bus, usart_rx_byte_callback);
    
    // При обнаружении свободной линии - прекратить принимать данные.
    usart_bus_set_idle_mode(&usart_bus, USART_IDLE_MODE_END_RX);
    
    // Разрешаем прерывания USART.
    NVIC_SetPriority(USART1_IRQn, 2);
    NVIC_EnableIRQ(USART1_IRQn);
    
    // Разрешаем прерывания DMA USART.
    NVIC_SetPriority(DMA1_Channel4_IRQn, 2);
    NVIC_EnableIRQ(DMA1_Channel4_IRQn);
    NVIC_SetPriority(DMA1_Channel5_IRQn, 2);
    NVIC_EnableIRQ(DMA1_Channel5_IRQn);
}

/*
 * Функции для работы с Modbus.
 */

/**
 * Обработчик получения сообщения Modbus.
 */
static void modbus_on_msg_recv(void)
{
    // Обработаем сообщение и перенаправим его в зависимости от функции.
    modbus_rtu_dispatch(&modbus);
}

/**
 * Обработчик чтения регистра флагов.
 * @param address Адрес.
 * @param value Значение.
 * @return Код ошибки Modbus.
 */
static modbus_rtu_error_t modbus_on_read_coil(uint16_t address, modbus_rtu_coil_value_t* value)
{
    // Если не адрес светодиода - возврат ошибки.
    if(address != LED_ADDRESS) return MODBUS_RTU_ERROR_INVALID_ADDRESS;

    // Передадим состояние светодиода.
    *value = led_status(LED0);
    
    return MODBUS_RTU_ERROR_NONE;
}

/**
 * Обработчик Записи регистра флагов.
 * @param address Адрес.
 * @param value Значение.
 * @return Код ошибки Modbus.
 */
static modbus_rtu_error_t modbus_on_write_coil(uint16_t address, modbus_rtu_coil_value_t value)
{
    // Если не адрес светодиода - возврат ошибки.
    if(address != LED_ADDRESS) return MODBUS_RTU_ERROR_INVALID_ADDRESS;
    
    // Зажжём или погасим светодиод
    // в зависимости от значения.
    if(value) led_on(LED0);
    else led_off(LED0);
    
    return MODBUS_RTU_ERROR_NONE;
}

/**
 * Обработчик чтения регистра хранения.
 * @param address Адрес.
 * @param value Значение.
 * @return Код ошибки Modbus.
 */
static modbus_rtu_error_t modbus_on_read_hold_reg(uint16_t address, uint16_t* value)
{
    // Если не адрес регистра счётчика - возврат ошибки.
    if(address != CNT_ADDRESS) return MODBUS_RTU_ERROR_INVALID_ADDRESS;
    
    // Передадим значение регистра счётчика.
    *value = cnt_reg;
    
    return MODBUS_RTU_ERROR_NONE;
}

/**
 * Обработчик записи регистра хранения.
 * @param address Адрес.
 * @param value Значение.
 * @return Код ошибки Modbus.
 */
static modbus_rtu_error_t modbus_on_write_reg(uint16_t address, uint16_t value)
{
    // Если не адрес регистра счётчика - возврат ошибки.
    if(address != CNT_ADDRESS) return MODBUS_RTU_ERROR_INVALID_ADDRESS;
    
    // Установим значение регистра счётчика.
    cnt_reg = value;
    
    return MODBUS_RTU_ERROR_NONE;
}

/**
 * Обработчик передачи идентификатора ведомого устройства.
 * @param slave_id Идентификатор ведомого устройства.
 * @return Код ошибки Modbus.
 */
modbus_rtu_error_t modbus_on_report_slave_id(modbus_rtu_slave_id_t* slave_id)
{
    // Состояние - работаем.
    slave_id->status = MODBUS_RTU_RUN_STATUS_ON;
    // Идентификатор - для пример возьмём 0xaa.
    slave_id->id = 0xaa;
    // В дополнительных данных передадим наше имя.
    slave_id->data = "STM32 MCU Modbus v1.0";
    // Длина имени.
    slave_id->data_size = 21;
    
    return MODBUS_RTU_ERROR_NONE;
}

/**
 * Инициализация Modbus.
 */
static void init_modbus(void)
{
    // Структура инициализации Modbus.
    modbus_rtu_init_t modbus_is;
    
    modbus_is.usart = &usart_bus; // Шина USART.
    modbus_is.mode = MODBUS_RTU_MODE_SLAVE; // Режим - ведомый.
    modbus_is.address = 0xaa; // Адрес.
    modbus_is.rx_message = &modbus_rx_msg; // Сообщение для приёма.
    modbus_is.tx_message = &modbus_tx_msg; // Сообщение для передачи.
    
    // Инициализируем Modbus.
    modbus_rtu_init(&modbus, &modbus_is);
    // Установка каллбэка получения сообщения.
    modbus_rtu_set_msg_recv_callback(&modbus, modbus_on_msg_recv);
    // Установка каллбэков доступа к данным.
    modbus_rtu_set_read_coil_callback(&modbus, modbus_on_read_coil);
    modbus_rtu_set_write_coil_callback(&modbus, modbus_on_write_coil);
    modbus_rtu_set_report_slave_id_callback(&modbus, modbus_on_report_slave_id);
    modbus_rtu_set_read_holding_reg_callback(&modbus, modbus_on_read_hold_reg);
    modbus_rtu_set_write_holding_reg_callback(&modbus, modbus_on_write_reg);
}

/**
 * Инициализирует системный счётчик.
 */
static void init_sys_counter(void)
{
    system_counter_init(1000);
    counter = system_counter_ticks();
    SysTick_Config(SystemCoreClock / 1000);
}

/**
 * Главная функция.
 */
int main(void)
{
    // Установка групп приоритета прерываний.
    NVIC_SetPriorityGrouping(0x3);
    // Инициализация.
    init_periph_clock();
    init_sys_counter();
    init_leds();
    init_usart();
    init_modbus();
    
    // Главный цикл.
    for(;;){
        // Если достчитали до секунды.
        if(system_counter_diff(&counter) >= system_counter_ticks_per_sec()){
            counter = system_counter_ticks();
            // Инкрементируем регистр счёта Modbus.
            cnt_reg ++;
        }
    }
    return 0;
}
